@@ -98,8 +98,10 @@
 
 (define *cload-cflags* (if (provided? 'clang) "-fPIC" ""))
 (define *cload-ldflags* "")
-(if (not (defined? '*cload-directory*))
-    (define *cload-directory* ""))
+(unless (defined? '*cload-directory*)
+  (define *cload-directory* ""))
+(unless (defined? '*cload-library-name*)
+  (define *cload-library-name* #f))
 
 (define *cload-c-compiler* ;(if (provided? 'tcc)  ; how to get tcc to write a .so file?
 			       ;"tcc"
@@ -129,7 +131,7 @@
 
   (define handlers (list '(integer s7_is_integer s7_integer s7_make_integer s7_int)
 			 '(boolean s7_is_boolean s7_boolean s7_make_boolean bool)
-			 '(real s7_is_real s7_number_to_real s7_make_real s7_double)
+			 '(real s7_is_real s7_number_to_real_with_caller s7_make_real s7_double)
 
 			 ;; '(complex s7_is_complex #f s7_make_complex s7_Complex)
 			 ;; the typedef is around line 6116 in s7.c, but we also need s7_complex which requires the s7_Complex type
@@ -288,7 +290,8 @@
 	     (lambda (header)
 	       (format p "#include <~A>~%" header))
 	     headers))
-	(format p "#include \"s7.h\"~%~%"))
+	(format p "#include \"s7.h\"~%~%")
+	(format p "static s7_pointer fsym, s7_F, s7_unspec, ffunc, c_pointer_string, string_string, character_string, boolean_string, real_string, complex_string, integer_string;~%"))
       
       (define collides?
 	(let ((all-names (hash-table)))
@@ -367,19 +370,29 @@
 				      (hyphen->space nominal-type)
 				      (type->type-symbol nominal-type)  ;(symbol->string nominal-type)
 				      (if (= num-args 1) 0 (+ i 1)))
-			      (begin
+
+			      (let ((function-name (if *cload-library-name*
+						       (string-append "(" *cload-library-name* " '" scheme-name ")")
+						       (if (> (length prefix) 0)
+							   scheme-name
+							   (substring base-name 4)))))
 				(format pp "  if (~A(arg))~%" (checker true-type))
-				(format pp "    ~A_~D = (~A)~A(~Aarg);~%"
+				(format pp "    ~A_~D = (~A)~A(~Aarg~A);~%"
 					base-name i
 					(hyphen->space nominal-type)
 					(s7->C true-type)                               ; s7_number_to_real which requires 
 					(if (memq s7-type '(boolean real))              ;   the extra sc arg
-					    "sc, " ""))
-				(format pp "  else return(s7_wrong_type_arg_error(sc, __func__, ~D, arg, ~S));~%"
+					    "sc, " 
+					    "")
+					(if (memq s7-type '(boolean real))              ;   the extra trailing caller arg
+					    ", __func__" 
+					    ""))
+				(format pp "  else return(s7_wrong_type_error(sc, s7_make_string_wrapper_with_length(sc, ~S, ~D), ~D, arg, ~A_string));~%"
+					function-name
+					(length function-name)
 					(if (= num-args 1) 0 (+ i 1))
-					(if (symbol? s7-type) 
-					    (symbol->string s7-type) 
-					    (error 'bad-arg (format #f "in ~S, ~S is not a symbol~%" name s7-type)))))))
+					s7-type))))
+
 		      (if (< i (- num-args 1))
 			  (format pp "  p = s7_cdr(p);~%")))))
 		
@@ -401,18 +414,19 @@
 		  (format pp ")")
 		  
 		  (if (eq? return-translator 's7_make_c_pointer_with_type)
-		      (format pp ", ~S, s7_f(sc))" (type->type-symbol return-type))
+		      (format pp ", ~S, s7_F)" (type->type-symbol return-type))
 		      (if (symbol? return-translator)
 			  (format pp ")")))
 		  (format pp (if (not (eq? return-translator #t))
 				 ");~%"
-				 ";~%  return(s7_unspecified(sc));~%"))
+				 ";~%  return(s7_unspec);~%"))
 		  (format pp "}~%"))
 		
 		;; add optimizer connection
 		(when (and (eq? return-type 'double)
 			   (< num-args 5)
-			   (sig-every? (lambda (p) (eq? p 'double)) arg-types))
+			   (sig-every? (lambda (p) (eq? p 'double)) arg-types)
+			   (not (defined? (symbol scheme-name) (rootlet)))) ; see below, double-funcs entry not used if already defined
 		  (let ((local-name #f))
 		    (case num-args
 		      ((0)
@@ -436,8 +450,9 @@
 		       (format pp "static s7_double ~A~A(s7_double x1, s7_double x2, s7_double x3, s7_double x4) {return(~A(x1, x2, x3, x4));}~%"
 			       func-name local-name func-name)))
 		    (set! double-funcs (cons (list func-name scheme-name local-name) double-funcs))))
-		
+
 		(when (and (memq return-type '(int size_t))        ; int (f int|double|void)
+			   (not (defined? (symbol scheme-name) (rootlet))) ; see below, int-funcs entry not used if already defined
 			   (or ;(= num-args 0)
 			    (and (= num-args 1)
 				 (memq (car arg-types) '(int size_t double)))
@@ -472,6 +487,7 @@
 		;; other possibilities: d_7pi|pii p=double* etc piid=checks in s7 (assumes float-vector)
 		;;   d_pd [lots of d_pdd, d_p, p_i and i_p]
 		;;   but how to recognize the "p" portions? (d_7pi with p="s7_pointer" gets no hits in libgsl)
+		;; libgsl: d_i: 15, d_ii: 5, [handled: d_id: 32], d_iid: 7, d_ddi: 4, d_idd: 10, d_idi: 6, i_ddd i_iii: 1, i_dddd: 4
 		
 		(format pp "~%")
 		(set! functions (cons (list scheme-name base-name 
@@ -545,6 +561,16 @@
 	   signatures)
 	  (format p "  }~%~%"))
 
+	(format p "  s7_F = s7_f(sc);~%")
+	(format p "  s7_unspec = s7_unspecified(sc);~%")
+	(format p "  string_string = s7_make_semipermanent_string(sc, \"a string\");~%")
+	(format p "  c_pointer_string = s7_make_semipermanent_string(sc, \"a c-pointer\");~%")
+	(format p "  character_string = s7_make_semipermanent_string(sc, \"a character\");~%")
+	(format p "  boolean_string = s7_make_semipermanent_string(sc, \"a boolean\");~%")
+	(format p "  real_string = s7_make_semipermanent_string(sc, \"a real\");~%")
+	(format p "  complex_string = s7_make_semipermanent_string(sc, \"a complex number\");~%")
+	(format p "  integer_string = s7_make_semipermanent_string(sc, \"an integer\");~%")
+
 	(format p "  cur_env = s7_curlet(sc);~%") ; changed from s7_outlet(s7_curlet) 20-Aug-17
 	
 	;; send out any special initialization code
@@ -570,7 +596,7 @@
 		    (scheme-name (string-append prefix (if (> (length prefix) 0) ":" "") c-name))
 		    (trans (C->s7 type)))
 	       (if (eq? trans 's7_make_c_pointer_with_type)
-		   (format p "  s7_define(sc, cur_env, s7_make_symbol(sc, ~S), ~A(sc, (~A)~A, s7_make_symbol(sc, ~S), s7_f(sc)));~%" 
+		   (format p "  s7_define(sc, cur_env, s7_make_symbol(sc, ~S), ~A(sc, (~A)~A, s7_make_symbol(sc, ~S), s7_F));~%" 
 			   scheme-name
 			   trans
 			   (C->s7-cast type)
@@ -594,7 +620,7 @@
 		    (trans (C->s7 type)))
 	       (format p "#ifdef ~A~%" c-name)
 	       (if (eq? trans 's7_make_c_pointer_with_type)
-		   (format p "  s7_define(sc, cur_env, s7_make_symbol(sc, ~S), s7_make_c_pointer_with_type(sc, (~A)~A, s7_make_symbol(sc, \"~S\"), s7_f(sc)));~%" 
+		   (format p "  s7_define(sc, cur_env, s7_make_symbol(sc, ~S), s7_make_c_pointer_with_type(sc, (~A)~A, s7_make_symbol(sc, \"~S\"), s7_F));~%" 
 			   scheme-name
 			   (C->s7-cast type)
 			   c-name
@@ -612,18 +638,19 @@
 	 (lambda (f) 
 	   (let ((scheme-name (f 0))
 		 (base-name   (f 1))
-		 (help        (f 2))
+		 (helpf       (f 2))
 		 (num-args    (f 3))
 		 (opt-args    (if (> (length f) 4) (f 4) 0))
 		 (sig         (and (> (length f) 5) (f 5))))
-	     (format p "~%  s7_define(sc, cur_env,~%            s7_make_symbol(sc, ~S),~%" scheme-name)
-	     (format p "            s7_make_typed_function(sc, ~S, ~A, ~D, ~D, false, ~S, ~A));~%"
+	     (format p "~%  s7_define(sc, cur_env,~%            fsym = s7_make_symbol(sc, ~S),~%" scheme-name)
+	     (format p "            ffunc = s7_make_typed_function_with_environment(sc, ~S, ~A, ~D, ~D, false, ~S, ~A, cur_env));~%"
 		     scheme-name
 		     base-name
 		     num-args
 		     opt-args
-		     help
-		     (if (pair? sig) (signatures sig) 'NULL))))
+		     helpf
+		     (if (pair? sig) (signatures sig) 'NULL))
+	     (format p "  s7_symbol_set_initial_value(sc, fsym, ffunc);~%"))) ; not sure this is a good idea, added 24-May-24
 	 functions)
 	
 	;; optimizer connection
